
#  ============LICENSE_START===============================================
#  Copyright (C) 2021 Nordix Foundation. All rights reserved.
#  ========================================================================
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#  ============LICENSE_END=================================================
#

from flask import Flask
from flask import request

import requests
import json
from jsonschema import validate
import threading
import time
from datetime import datetime
import logging
import os
import sys
import traceback

from requests.packages import urllib3

# Remove warning fot https calls
urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

app = Flask(__name__)

#Set to True to log att exception traces
debug=False

# Server info
HOST_IP = "::"
HOST_PORT = 2222

datatypes = {}
registration_status=0 # 0 nothing registered, 1 types registred, 2 producer registered


PUT_PRODUCER="/ei-producer/v1/eiproducers/"

# Request and response constants
CALLBACK_CREATE_BASE="/callbacks/job"
CALLBACK_CREATE_URL=CALLBACK_CREATE_BASE+"/<string:producer_id>"

CALLBACK_DELETE_BASE=CALLBACK_CREATE_BASE
CALLBACK_DELETE_URL=CALLBACK_CREATE_URL+"/<string:job_id>"

CALLBACK_SUPERVISION_BASE="/callbacks/supervision"
CALLBACK_SUPERVISION_URL=CALLBACK_SUPERVISION_BASE+"/<string:producer_id>"

STATUS="/status"
RESET="/reset"

#Constsants
JOB_NOT_FOUND="job not found"
JSON_CORRUPT="json in request is corrupt or missing"

#Producer and job db, including armed responses
db={}

# Get status info
# URI and parameters (GET): "/status"
# -
@app.route(STATUS,
    methods=['GET'])
def status():
    global db
    return json.dumps(db),200


# Reset db
@app.route(RESET,
    methods=['POST'])
def reset():
    global db
    db={}
    registration_status=0
    return "",200

# Callback for create job
# URI and parameters (POST): /callbacks/job/<producer-id>
# response 201 at create, 200 at update or other configured response code
@app.route(CALLBACK_CREATE_URL,
     methods=['POST'])
def callback_create(producer_id):

    try:
        req_json = json.loads(request.data)
        print("Callback for create, producer: " + producer_id + " json: " + json.dumps(req_json))
        job_type=req_json['ei_type_identity']
        if job_type in datatypes.keys():
            datatype=datatypes[job_type]
            validate(instance=req_json, schema=datatype)
            job_id=req_json['ei_job_identity']
            print("Job "+ job_id + " created with type "+job_type+" requested")
            db[job_id]=req_json
    except Exception:
        if debug:
            traceback.print_exc()
        return JSON_CORRUPT,400

    return "OK", 200


# Callback for delete job
# URI and parameters (DELETE): /callbacks/job/<producer-id>/<job_id>
# response: 204 at delete or other configured response code
@app.route(CALLBACK_DELETE_URL,
     methods=['DELETE'])
def callback_delete(producer_id, job_id):

    print("Callback for delete, producer: " + producer_id + " and job: " + job_id)

    if job_id not in db.keys():
        return JOB_NOT_FOUND,400
    del db[job_id]

    return "", 204

# Callback for supervision of producer
# URI and parameters (GET): /callbacks/supervision/<producer-id>
# response: 200 or other configured response code
@app.route(CALLBACK_SUPERVISION_URL,
     methods=['GET'])
def callback_supervision(producer_id):

    print("Supervision callback received")
    return "OK",200


# Create the job towards ecs
def handle_jobs(param_dict):

    print("Parameters in handle jobs: "+ json.dumps(param_dict))
    headers = {'content-type': 'application/json'}

    for key in db:
        job_result={}
        job_result["date"]=str(datetime.now())
        job_result["id"]=db[key]["ei_job_identity"]
        pnames=db[key]["ei_job_data"]["parameteter_names"]
        print(json.dumps(pnames))
        for pn in pnames:
            if pn in param_dict.keys():
                job_result[pn]=param_dict[pn]["value"]

        print("job_result for job " + db[key]["ei_job_identity"] + ":"+json.dumps(job_result))
        if len(job_result)>0:
            try:
                url=db[key]["target_uri"]
                print("Sending job " + db[key]["ei_job_identity"] + " to "+url)
                resp=requests.post(url, json=job_result, headers=headers, verify=False, timeout=2)
                if resp.status_code == 200:
                    print("Post for job " + job_result["id"] +" OK, status: "+str(resp.status_code))
                else:
                    print("Post for job " + job_result["id"] +" failed, status: "+str(resp.status_code))
            except Exception as err:
                traceback.print_exc()
                print("Error during job handling of job " + job_result["id"]  + ": "+ str(err))
        else:
            print("job_result for job " + job_result["id"] + " is empty, nothing is sent")

#======================================Available params in the forecast data======================================#
#name      unit      level-type  level  description                              data type
#msl       hPa       hmsl         0     Air pressure                             Decimal number, one decimal
#t         C         hl           2     Air temperature                          Decimal number, one decimal
#vis       km        hl           2     Horizontal visibility                    Decimal number, one decimal
#wd        degree    hl          10     Wind direction                           Integer
#ws        m/s       hl          10     Wind speed                               Decimal number, one decimal
#r         %         hl           2     Relative humidity                        Integer, 0-100
#tstm      %         hl           0     Thunder probability                      Integer, 0-100
#tcc_mean  octas     hl           0     Mean value of total cloud cover          Integer, 0-8
#lcc_mean  octas     hl           0     Mean value of low level cloud cover      Integer, 0-8
#mcc_mean  octas     hl           0     Mean value of medium level cloud cover   Integer, 0-8
#hcc_mean  octas     hl           0     Mean value of high level cloud cover     Integer, 0-8
#gust      m/s       hl          10     Wind gust speed                          Decimal number, one decimal
#pmin      mm/h      hl           0     Minimum precipitation intensity          Decimal number, one decimal
#pmax      mm/h      hl           0     Maximum precipitation intensity          Decimal number, one decimal
#spp       %         hl           0     Percent of precipitation in frozen form  Integer, -9 or 0-100
#pcat      category  hl           0     Precipitation category                   Integer, 0-6
#pmean     mm/h      hl           0     Mean precipitation intensity             Decimal number, one decimal
#pmedian   mm/h      hl           0     Median precipitation intensity           Decimal number, one decimal
#=================================================================================================================#

# Get data from external source
def data_controller():
    global registered
    global datatypes

    print("Data controller started")

    url="https://opendata-download-metfcst.smhi.se/api/category/pmp3g/version/2/geotype/point/lon/16.158/lat/58.5812/data.json"

    sleep_time=1
    while True:
        if len(db) > 0:
            sleep_time=60
            try:
                resp=requests.get(url, verify=False, timeout=2)
                if resp.status_code == 200:

                    req_json = json.loads(resp.text)
                    params=req_json["timeSeries"][0]["parameters"]
                    param_dict={}
                    for pn in params:
                        pname=pn["name"]
                        pval=pn["values"][0]
                        if (pname=="t") or (pname=="vis") or (pname=="ws") or (pname=="wd") or (pname=="gust"):
                            tmp={}
                            tmp["parameter"]=pname
                            tmp["value"]=pval
                            param_dict[pname]=tmp
                    if len(param_dict) > 0:
                        handle_jobs(param_dict)

            except Exception as err:
                if debug:
                    traceback.print_exc()
                print("Error during data handling: "+ str(err))
                sleep_time=10
        else:
            sleep_time=1

        time.sleep(sleep_time)

# Read data types from disk
def read_data_types():
    # Read data types file
    print("Load datatypes...")
    path_to_jsonfiles="./job-data-types"
    for file in os.listdir(path_to_jsonfiles):
        full_filename = "%s/%s" % (path_to_jsonfiles, file)
        if full_filename.endswith(".json"):
            print("Loading data type file: "+full_filename)
            try:
                with open(full_filename,'r') as fi:
                    type_name=str(os.path.splitext(file)[0])
                    print(" type name: "+type_name)
                    file_dict = json.load(fi)
                    datatypes[type_name]=file_dict
            except Exception as err:
                if debug:
                    traceback.print_exc()
                print("File corrupt")

# Register type and producer
def producer_registration() :
    global registration_status
    global datatypes

    print("Registration controller started")

    read_data_types()

    headers = {'content-type': 'application/json'}

    while True:
        try:
            if registration_status == 0:
                num_types=len(datatypes)
                type_ctr=0
                for key in datatypes:
                    json_type={}
                    json_type["ei_job_data_schema"]=datatypes[key]
                    url=ecs_base+"/ei-producer/v1/eitypes/"+key
                    print("Register type "+key+" on "+ url)
                    print("Data type: " + json.dumps(json_type))
                    resp=requests.put(url, json=json_type, headers=headers, verify=False, timeout=2)
                    print(resp.text)
                    if resp.status_code == 200:
                        print("Type "+key+ " updated - status: "+str(resp.status_code))
                        type_ctr=type_ctr+1
                    elif resp.status_code == 201:
                        print("Type "+key+ " created - status: "+str(resp.status_code))
                        type_ctr=type_ctr+1
                    else:
                        print("Registration of type " + key + " failed, status: "+str(resp.status_code))

                if (num_types > 0 and num_types == type_ctr):
                    print("All types registered")
                    registration_status=1
                elif (num_types == 0):
                    print("No types found to register")
                    registration_status=1

            if registration_status == 1:
                print("Register producer "+producer_name)
                producer_data={}
                type_data=[]
                for key in datatypes:
                    type_data.append(key)
                producer_data["supported_ei_types"]=type_data
                producer_data["ei_job_callback_url"]=own_host+CALLBACK_CREATE_BASE+"/"+producer_name
                producer_data["ei_producer_supervision_callback_url"]=own_host+CALLBACK_SUPERVISION_BASE+"/"+producer_name

                url=ecs_base+"/ei-producer/v1/eiproducers/"+producer_name
                print("Register producer "+producer_name+" on "+ url)
                print("Producer json: " + json.dumps(producer_data))
                resp=requests.put(url, json=producer_data, headers=headers, verify=False, timeout=2)
                if resp.status_code == 200:
                    print("Updated producer: "+str(resp.status_code))
                    registration_status=2
                elif resp.status_code == 201:
                    print("Registered producer: "+str(resp.status_code))
                    registration_status=2
                else:
                    print("Registration of producer failed, status: "+str(resp.status_code))

        except Exception as err:
            if debug:
                traceback.print_exc()
            print("Error during type or producer registration: "+ str(err))
        time.sleep(3)


### Main function ###

# Set host port - only for local python execution from the cmd line

if len(sys.argv) > 1 :
    HOST_PORT=int(sys.argv[1])

ecs_base=os.getenv('ECS_BASE')
if ecs_base is None:
    sys.exit("Enviroment var ECS_BASE not set")
else:
    print("Enviroment var ECS_BASE = "+ecs_base)

own_host=os.getenv('OWN_HOST')
if own_host is None:
    sys.exit("Enviroment var OWN_HOST not set")
else:
    print("Enviroment var OWN_HOST = "+own_host)

producer_name=os.getenv('PRODUCER_NAME')
if producer_name is None:
    sys.exit("Enviroment var PRODUCER_NAME not set")
else:
    print("Enviroment var PRODUCER_NAME = "+producer_name)

registered=False

thread_reg = threading.Thread(target=producer_registration, args=())
thread_reg.daemon = True
thread_reg.start()

thread_data = threading.Thread(target=data_controller, args=())
thread_data.daemon = True
thread_data.start()

if __name__ == "__main__":
    print("app.run on port: "+ str(HOST_PORT))
    app.run(port=HOST_PORT, host=HOST_IP)
